@@ -4,6 +4,8 @@ Defines a scalar field over a grid
 .. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
 """
 
+from __future__ import annotations
+
 import numbers
 from pathlib import Path
 from typing import List  # @UnusedImport
@@ -23,16 +25,7 @@ if TYPE_CHECKING:
 
 
 class ScalarField(DataFieldBase):
-    """Single scalar field on a grid
-
-    Attributes:
-        grid (:class:`~pde.grids.base.GridBase`):
-            The underlying grid defining the discretization
-        data (:class:`np.ndarray`):
-            Scalar values at the support points of the grid
-        label (str):
-            Name of the field
-    """
+    """Scalar field discretized on a grid"""
 
     rank = 0
 
@@ -40,7 +33,7 @@ class ScalarField(DataFieldBase):
     @fill_in_docstring
     def from_expression(
         cls, grid: GridBase, expression: str, *, label: str = None, dtype=None
-    ) -> "ScalarField":
+    ) -> ScalarField:
         """create a scalar field on a grid from a given expression
 
         Warning:
@@ -71,7 +64,7 @@ class ScalarField(DataFieldBase):
     @classmethod
     def from_image(
         cls, path: Union[Path, str], bounds=None, periodic=False, *, label: str = None
-    ) -> "ScalarField":
+    ) -> ScalarField:
         """create a scalar field from an image
 
         Args:
@@ -113,7 +106,7 @@ class ScalarField(DataFieldBase):
     @DataFieldBase._data_flat.setter  # type: ignore
     def _data_flat(self, value):
         """set the data from a value from a collection"""
-        self._data = value[0]
+        self._data_all = value[0]
 
     def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
         """support unary numpy ufuncs, like np.sin, but also np.multiply"""
@@ -151,12 +144,8 @@ class ScalarField(DataFieldBase):
 
     @fill_in_docstring
     def laplace(
-        self,
-        bc: "BoundariesData",
-        out: Optional["ScalarField"] = None,
-        *,
-        label: str = "laplace",
-    ) -> "ScalarField":
+        self, bc: "BoundariesData", out: Optional[ScalarField] = None, **kwargs
+    ) -> ScalarField:
         """apply Laplace operator and return result as a field
 
         Args:
@@ -167,22 +156,18 @@ class ScalarField(DataFieldBase):
                 Optional scalar field to which the  result is written.
             label (str, optional):
                 Name of the returned field
+            backend (str):
+                The backend (e.g., 'numba' or 'scipy') used for this operator.
 
         Returns:
             :class:`~pde.fields.scalar.ScalarField`: the Laplacian of the field
         """
-        laplace = self.grid.get_operator("laplace", bc=bc)
-        return self._apply_with_out(laplace, ScalarField, out=out, label=label)
+        return self._apply_operator("laplace", bc=bc, out=out, **kwargs)  # type: ignore
 
     @fill_in_docstring
     def gradient_squared(
-        self,
-        bc: "BoundariesData",
-        out: Optional["ScalarField"] = None,
-        *,
-        central: bool = True,
-        label: str = "squared gradient",
-    ) -> "ScalarField":
+        self, bc: "BoundariesData", out: Optional[ScalarField] = None, **kwargs
+    ) -> ScalarField:
         r"""apply squared gradient operator and return result as a field
 
         This evaluates :math:`|\nabla \phi|^2` for the scalar field :math:`\phi`
@@ -191,31 +176,25 @@ class ScalarField(DataFieldBase):
             bc:
                 The boundary conditions applied to the field.
                 {ARG_BOUNDARIES}
-            central (bool):
-                Determines whether a central difference approximation is used
-                for the gradient operator or not. If not, the squared gradient
-                is calculated as the mean of the squared values of the forward
-                and backward derivatives, which thus includes the value at a
-                support point in the result at the same point.
             out (ScalarField, optional):
-                Optional scalar field to which the  result is written.
+                Optional vector field to which the result is written.
             label (str, optional):
                 Name of the returned field
+            central (bool):
+                Determines whether a central difference approximation is used for the
+                gradient operator or not. If not, the squared gradient is calculated as
+                the mean of the squared values of the forward and backward derivatives,
+                which thus includes the value at a support point in the result at the
+                same point.
 
         Returns:
             :class:`~pde.fields.scalar.ScalarField`: the squared gradient of the field
         """
-
-        grad_square = self.grid.get_operator("gradient_squared", bc=bc, central=central)
-        return self._apply_with_out(grad_square, ScalarField, out=out, label=label)
+        return self._apply_operator("gradient_squared", bc=bc, out=out, **kwargs)  # type: ignore
 
     @fill_in_docstring
     def gradient(
-        self,
-        bc: "BoundariesData",
-        out: Optional["VectorField"] = None,
-        *,
-        label: str = "gradient",
+        self, bc: "BoundariesData", out: Optional["VectorField"] = None, **kwargs
     ) -> "VectorField":
         """apply gradient operator and return result as a field
 
@@ -231,10 +210,7 @@ class ScalarField(DataFieldBase):
         Returns:
             :class:`~pde.fields.vectorial.VectorField`: result of applying the operator
         """
-        from .vectorial import VectorField  # @Reimport
-
-        gradient = self.grid.get_operator("gradient", bc=bc)
-        return self._apply_with_out(gradient, VectorField, out=out, label=label)
+        return self._apply_operator("gradient", bc=bc, out=out, **kwargs)  # type: ignore
 
     @property
     def integral(self) -> Number:
@@ -246,7 +222,7 @@ class ScalarField(DataFieldBase):
         axes: Union[str, Sequence[str]],
         method: str = "integral",
         label: str = None,
-    ) -> "ScalarField":
+    ) -> ScalarField:
         """project scalar field along given axes
 
         Args:
@@ -298,7 +274,7 @@ class ScalarField(DataFieldBase):
 
     def slice(
         self, position: Dict[str, float], *, method: str = "nearest", label: str = None
-    ) -> "ScalarField":
+    ) -> ScalarField:
         """slice data at a given position
 
         Args:
@@ -378,7 +354,7 @@ class ScalarField(DataFieldBase):
 
     def to_scalar(
         self, scalar: Union[str, Callable] = "auto", *, label: Optional[str] = None
-    ) -> "ScalarField":
+    ) -> ScalarField:
         """return a modified scalar field by applying `method`
 
         Args:
