@@ -19,7 +19,7 @@ from typing import Tuple
 import numpy as np
 
 from ...tools.docstrings import fill_in_docstring
-from ...tools.numba import jit_allocate_out
+from ...tools.numba import jit
 from ...tools.typing import OperatorType
 from ..boundaries import Boundaries
 from ..spherical import SphericalSymGrid
@@ -28,155 +28,100 @@ from .common import make_general_poisson_solver
 
 @SphericalSymGrid.register_operator("laplace", rank_in=0, rank_out=0)
 @fill_in_docstring
-def make_laplace(bcs: Boundaries, conservative: bool = True) -> OperatorType:
+def make_laplace(grid: SphericalSymGrid, conservative: bool = True) -> OperatorType:
     """make a discretized laplace operator for a spherical grid
 
     {DESCR_SPHERICAL_GRID}
 
     Args:
-        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
-            {ARG_BOUNDARIES_INSTANCE}
-        conservative (bool): flag indicating whether the laplace operator should
-            be conservative (which results in slightly slower computations).
+        grid (:class:`~pde.grids.spherical.SphericalSymGrid`):
+            The polar grid for which this operator will be defined
+        conservative (bool):
+            Flag indicating whether the laplace operator should be conservative (which
+            results in slightly slower computations).
 
     Returns:
         A function that can be applied to an array of values
     """
-    assert isinstance(bcs.grid, SphericalSymGrid)
-    bcs.check_value_rank(0)
+    assert isinstance(grid, SphericalSymGrid)
 
     # calculate preliminary quantities
-    dim_r = bcs.grid.shape[0]
-    dr = bcs.grid.discretization[0]
-    rs = bcs.grid.axes_coords[0]
-    r_min, r_max = bcs.grid.axes_bounds[0]
-
-    # prepare boundary values
-    value_lower_bc = bcs[0].low.make_virtual_point_evaluator()
-    value_upper_bc = bcs[0].high.make_virtual_point_evaluator()
+    dim_r = grid.shape[0]
+    dr = grid.discretization[0]
+    rs = grid.axes_coords[0]
+    r_min, r_max = grid.axes_bounds[0]
 
     if conservative:
         # create a conservative spherical laplace operator
-        rl = r_min + dr * np.arange(dim_r)  # inner radii of spherical shells
-        rh = rl + dr  # outer radii
-        assert np.isclose(rh[-1], r_max)
+        rl = rs - dr / 2  # inner radii of spherical shells
+        rh = rs + dr / 2  # outer radii
+        assert np.isclose(rl[0], r_min) and np.isclose(rh[-1], r_max)
         volumes = (rh ** 3 - rl ** 3) / 3  # volume of the spherical shells
         factor_l = (rs - 0.5 * dr) ** 2 / (dr * volumes)
         factor_h = (rs + 0.5 * dr) ** 2 / (dr * volumes)
 
-        @jit_allocate_out(out_shape=(dim_r,))
-        def laplace(arr, out=None):
+        @jit
+        def laplace(arr: np.ndarray, out: np.ndarray) -> None:
             """apply laplace operator to array `arr`"""
-            i = 0
-            out[i] = factor_h[i] * (arr[i + 1] - arr[i])
-            if r_min > 0:
-                arr_r_l = value_lower_bc(arr, (i,))
-                out[i] -= factor_l[i] * (arr[i] - arr_r_l)
-
-            for i in range(1, dim_r - 1):  # iterate inner radial points
-                out[i] = factor_h[i] * (arr[i + 1] - arr[i])
-                out[i] -= factor_l[i] * (arr[i] - arr[i - 1])
-
-            # express boundary condition at outer side
-            i = dim_r - 1
-            arr_r_h = value_upper_bc(arr, (i,))
-            out[i] = factor_h[i] * (arr_r_h - arr[i])
-            out[i] -= factor_l[i] * (arr[i] - arr[i - 1])
-            return out
+            for i in range(1, dim_r + 1):  # iterate inner radial points
+                out[i] = factor_h[i - 1] * (arr[i + 1] - arr[i])
+                out[i] -= factor_l[i - 1] * (arr[i] - arr[i - 1])
 
     else:  # create an operator that is not conservative
         dr2 = 1 / dr ** 2
 
-        @jit_allocate_out(out_shape=(dim_r,))
-        def laplace(arr, out=None):
+        @jit
+        def laplace(arr: np.ndarray, out: np.ndarray) -> None:
             """apply laplace operator to array `arr`"""
-            i = 0
-            if r_min == 0:
-                out[i] = 3 * (arr[i + 1] - arr[i]) * dr2
-            else:
-                arr_r_l = value_lower_bc(arr, (i,))
-                out[i] = (arr[i + 1] - 2 * arr[i] + arr_r_l) * dr2
-                out[i] += (arr[i + 1] - arr_r_l) / (rs[i] * dr)
-
-            for i in range(1, dim_r - 1):  # iterate inner radial points
+            for i in range(1, dim_r + 1):  # iterate inner radial points
                 out[i] = (arr[i + 1] - 2 * arr[i] + arr[i - 1]) * dr2
-                out[i] += (arr[i + 1] - arr[i - 1]) / (rs[i] * dr)
-
-            # express boundary condition at outer side
-            i = dim_r - 1
-            arr_r_h = value_upper_bc(arr, (i,))
-            out[i] = (arr_r_h - 2 * arr[i] + arr[i - 1]) * dr2
-            out[i] += (arr_r_h - arr[i - 1]) / (rs[i] * dr)
-            return out
+                out[i] += (arr[i + 1] - arr[i - 1]) / (rs[i - 1] * dr)
 
     return laplace  # type: ignore
 
 
 @SphericalSymGrid.register_operator("gradient", rank_in=0, rank_out=1)
 @fill_in_docstring
-def make_gradient(bcs: Boundaries) -> OperatorType:
+def make_gradient(grid: SphericalSymGrid) -> OperatorType:
     """make a discretized gradient operator for a spherical grid
 
     {DESCR_SPHERICAL_GRID}
 
     Args:
-        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
-            {ARG_BOUNDARIES_INSTANCE}
+        grid (:class:`~pde.grids.spherical.SphericalSymGrid`):
+            The polar grid for which this operator will be defined
 
     Returns:
         A function that can be applied to an array of values
     """
-    assert isinstance(bcs.grid, SphericalSymGrid)
-    bcs.check_value_rank(0)
+    assert isinstance(grid, SphericalSymGrid)
 
     # calculate preliminary quantities
-    dim_r = bcs.grid.shape[0]
-    dr = bcs.grid.discretization[0]
-    r_min, _ = bcs.grid.axes_bounds[0]
+    dim_r = grid.shape[0]
+    dr = grid.discretization[0]
 
     scale_r = 1 / (2 * dr)
 
-    # prepare boundary values
-    boundary = bcs[0]
-    value_lower_bc = boundary.low.make_virtual_point_evaluator()
-    value_upper_bc = boundary.high.make_virtual_point_evaluator()
-
-    @jit_allocate_out(out_shape=(3, dim_r))
-    def gradient(arr, out=None):
+    @jit
+    def gradient(arr: np.ndarray, out: np.ndarray) -> None:
         """apply gradient operator to array `arr`"""
-        i = 0
-        if r_min == 0:
-            # Apply Neumann condition at the origin
-            out[0, i] = (arr[1] - arr[0]) * scale_r
-        else:
-            arr_r_l = value_lower_bc(arr, (i,))
-            out[0, i] = (arr[1] - arr_r_l) * scale_r
-        out[1, i] = out[2, i] = 0  # no angular dependence by definition
-
-        for i in range(1, dim_r - 1):  # iterate inner radial points
+        for i in range(1, dim_r + 1):  # iterate inner radial points
             out[0, i] = (arr[i + 1] - arr[i - 1]) * scale_r
             out[1, i] = out[2, i] = 0  # no angular dependence by definition
 
-        i = dim_r - 1
-        arr_r_h = value_upper_bc(arr, (i,))
-        out[0, i] = (arr_r_h - arr[i - 1]) * scale_r
-        out[1, i] = out[2, i] = 0  # no angular dependence by definition
-
-        return out
-
     return gradient  # type: ignore
 
 
 @SphericalSymGrid.register_operator("gradient_squared", rank_in=0, rank_out=0)
 @fill_in_docstring
-def make_gradient_squared(bcs: Boundaries, central: bool = True) -> OperatorType:
+def make_gradient_squared(grid: SphericalSymGrid, central: bool = True) -> OperatorType:
     """make a discretized gradient squared operator for a spherical grid
 
     {DESCR_SPHERICAL_GRID}
 
     Args:
-        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
-            {ARG_BOUNDARIES_INSTANCE}
+        grid (:class:`~pde.grids.spherical.SphericalSymGrid`):
+            The polar grid for which this operator will be defined
         central (bool):
             Whether a central difference approximation is used for the gradient
             operator. If this is False, the squared gradient is calculated as
@@ -186,72 +131,39 @@ def make_gradient_squared(bcs: Boundaries, central: bool = True) -> OperatorType
     Returns:
         A function that can be applied to an array of values
     """
-    assert isinstance(bcs.grid, SphericalSymGrid)
-    bcs.check_value_rank(0)
+    assert isinstance(grid, SphericalSymGrid)
 
     # calculate preliminary quantities
-    dim_r = bcs.grid.shape[0]
-    r_min, _ = bcs.grid.axes_bounds[0]
-    dr = bcs.grid.discretization[0]
-
-    # prepare boundary values
-    boundary = bcs[0]
-    value_lower_bc = boundary.low.make_virtual_point_evaluator()
-    value_upper_bc = boundary.high.make_virtual_point_evaluator()
+    dim_r = grid.shape[0]
+    dr = grid.discretization[0]
 
     if central:
         # use central differences
-        scale = 1 / (2 * dr) ** 2
+        scale = 0.25 / dr ** 2
 
-        @jit_allocate_out(out_shape=(dim_r,))
-        def gradient_squared(arr, out=None):
+        @jit
+        def gradient_squared(arr: np.ndarray, out: np.ndarray) -> None:
             """apply squared gradient operator to array `arr`"""
-            if r_min == 0:
-                # Apply Neumann condition at the origin
-                out[0] = (arr[1] - arr[0]) ** 2 * scale
-            else:
-                arr_r_l = value_lower_bc(arr, (0,))
-                out[0] = (arr[1] - arr_r_l) ** 2 * scale
-
-            for i in range(1, dim_r - 1):  # iterate inner radial points
+            for i in range(1, dim_r + 1):  # iterate inner radial points
                 out[i] = (arr[i + 1] - arr[i - 1]) ** 2 * scale
 
-            i = dim_r - 1
-            arr_r_h = value_upper_bc(arr, (i,))
-            out[i] = (arr_r_h - arr[i - 1]) ** 2 * scale
-
-            return out
-
     else:
         # use forward and backward differences
-        scale = 1 / (2 * dr ** 2)
+        scale = 0.5 / dr ** 2
 
-        @jit_allocate_out(out_shape=(dim_r,))
-        def gradient_squared(arr, out=None):
+        @jit
+        def gradient_squared(arr: np.ndarray, out: np.ndarray) -> None:
             """apply squared gradient operator to array `arr`"""
-            if r_min == 0:
-                # Apply Neumann condition at the origin
-                out[0] = (arr[1] - arr[0]) ** 2 * scale
-            else:
-                arr_r_l = value_lower_bc(arr, (0,))
-                out[0] = ((arr[1] - arr[0]) ** 2 + (arr[0] - arr_r_l) ** 2) * scale
-
-            for i in range(1, dim_r - 1):  # iterate inner radial points
+            for i in range(1, dim_r + 1):  # iterate inner radial points
                 term = (arr[i + 1] - arr[i]) ** 2 + (arr[i] - arr[i - 1]) ** 2
                 out[i] = term * scale
 
-            i = dim_r - 1
-            arr_r_h = value_upper_bc(arr, (i,))
-            out[i] = ((arr_r_h - arr[i]) ** 2 + (arr[i] - arr[i - 1]) ** 2) * scale
-
-            return out
-
     return gradient_squared  # type: ignore
 
 
 @SphericalSymGrid.register_operator("divergence", rank_in=1, rank_out=0)
 @fill_in_docstring
-def make_divergence(bcs: Boundaries, safe: bool = True) -> OperatorType:
+def make_divergence(grid: SphericalSymGrid, safe: bool = True) -> OperatorType:
     """make a discretized divergence operator for a spherical grid
 
     {DESCR_SPHERICAL_GRID}
@@ -262,83 +174,39 @@ def make_divergence(bcs: Boundaries, safe: bool = True) -> OperatorType:
         on a :class:`~pde.grids.spherical_sym.SphericalSymGrid`.
 
     Args:
-        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
-            {ARG_BOUNDARIES_INSTANCE}
+        grid (:class:`~pde.grids.spherical.SphericalSymGrid`):
+            The polar grid for which this operator will be defined
         safe (bool):
             Add extra checks for the validity of the input
 
     Returns:
         A function that can be applied to an array of values
     """
-    assert isinstance(bcs.grid, SphericalSymGrid)
-    bcs.check_value_rank(0)
+    assert isinstance(grid, SphericalSymGrid)
 
     # calculate preliminary quantities
-    dim_r = bcs.grid.shape[0]
-    dr = bcs.grid.discretization[0]
-    rs = bcs.grid.axes_coords[0]
-    r_min, _ = bcs.grid.axes_bounds[0]
+    dim_r = grid.shape[0]
+    dr = grid.discretization[0]
+    rs = grid.axes_coords[0]
 
     scale_r = 1 / (2 * dr)
+    fs = 2 / rs  # factors that need to be multiplied below
 
-    # prepare boundary values
-    boundary = bcs[0]
-    value_lower_bc = boundary.low.make_virtual_point_evaluator()
-    value_upper_bc = boundary.high.make_virtual_point_evaluator()
-
-    if r_min == 0:
-
-        @jit_allocate_out(out_shape=(dim_r,))
-        def divergence(arr, out=None):
-            """apply divergence operator to array `arr`"""
-            if safe:
-                assert np.all(arr[1, :] == 0)
-            arr_r = arr[0, :]
-            i = 0
-            out[i] = (arr_r[1] + 7 * arr_r[0]) * scale_r
-
-            for i in range(1, dim_r - 1):  # iterate inner radial points
-                out[i] = (arr_r[i + 1] - arr_r[i - 1]) * scale_r
-                out[i] += 2 * arr_r[i] / ((i + 0.5) * dr)
-
-            i = dim_r - 1
-            arr_r_h = value_upper_bc(arr[0], (i,))
-            out[i] = (arr_r_h - arr_r[i - 1]) * scale_r
-            out[i] += 2 * arr_r[i] / ((i + 0.5) * dr)
-
-            return out
-
-    else:  # r_min > 0
-        fs = 2 / rs  # factors that need to be multiplied below
-
-        @jit_allocate_out(out_shape=(dim_r,))
-        def divergence(arr, out=None):
-            """apply divergence operator to array `arr`"""
-            if safe:
-                assert np.all(arr[1, :] == 0)
-            arr_r = arr[0, :]
-
-            # inner radial boundary condition
-            i = 0
-            arr_r_l = value_lower_bc(arr[0], (i,))
-            out[i] = (arr_r[i + 1] - arr_r_l) * scale_r + fs[i] * arr_r[i]
-
-            for i in range(1, dim_r - 1):  # iterate radial points
-                out[i] = (arr_r[i + 1] - arr_r[i - 1]) * scale_r + fs[i] * arr_r[i]
-
-            # outer radial boundary condition
-            i = dim_r - 1
-            arr_r_h = value_upper_bc(arr[0], (i,))
-            out[i] = (arr_r_h - arr_r[i - 1]) * scale_r + fs[i] * arr_r[i]
-
-            return out
+    @jit
+    def divergence(arr: np.ndarray, out: np.ndarray) -> None:
+        """apply divergence operator to array `arr`"""
+        if safe:
+            assert np.all(arr[1, 1:-1] == 0)
+        arr_r = arr[0, :]
+        for i in range(1, dim_r + 1):  # iterate radial points
+            out[i] = (arr_r[i + 1] - arr_r[i - 1]) * scale_r + fs[i - 1] * arr_r[i]
 
     return divergence  # type: ignore
 
 
 @SphericalSymGrid.register_operator("vector_gradient", rank_in=1, rank_out=2)
 @fill_in_docstring
-def make_vector_gradient(bcs: Boundaries, safe: bool = True) -> OperatorType:
+def make_vector_gradient(grid: SphericalSymGrid, safe: bool = True) -> OperatorType:
     """make a discretized vector gradient operator for a spherical grid
 
     Warning:
@@ -349,34 +217,27 @@ def make_vector_gradient(bcs: Boundaries, safe: bool = True) -> OperatorType:
     {DESCR_SPHERICAL_GRID}
 
     Args:
-        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
-            {ARG_BOUNDARIES_INSTANCE}
+        grid (:class:`~pde.grids.spherical.SphericalSymGrid`):
+            The polar grid for which this operator will be defined
         safe (bool):
             Add extra checks for the validity of the input
 
     Returns:
         A function that can be applied to an array of values
     """
-    assert isinstance(bcs.grid, SphericalSymGrid)
-    bcs.check_value_rank(1)
+    assert isinstance(grid, SphericalSymGrid)
 
     # calculate preliminary quantities
-    dim_r = bcs.grid.shape[0]
-    r_min, _ = bcs.grid.axes_bounds[0]
-    rs = bcs.grid.axes_coords[0]
-    dr = bcs.grid.discretization[0]
+    dim_r = grid.shape[0]
+    rs = grid.axes_coords[0]
+    dr = grid.discretization[0]
     scale_r = 1 / (2 * dr)
 
-    # prepare boundary evaluators
-    bc_r = bcs.extract_component(0)[0]
-    value_r_lower_bc = bc_r.low.make_virtual_point_evaluator()
-    value_r_upper_bc = bc_r.high.make_virtual_point_evaluator()
-
-    @jit_allocate_out(out_shape=(3, 3, dim_r))
-    def vector_gradient(arr, out=None):
+    @jit
+    def vector_gradient(arr: np.ndarray, out: np.ndarray) -> None:
         """apply vector gradient operator to array `arr`"""
         if safe:
-            assert np.all(arr[1:, :] == 0)
+            assert np.all(arr[1:, 1:-1] == 0)
 
         # assign aliases
         arr_r = arr[0, :]
@@ -393,72 +254,40 @@ def make_vector_gradient(bcs: Boundaries, safe: bool = True) -> OperatorType:
         out_φθ[:] = 0
 
         # inner radial boundary condition
-        i = 0
-        if r_min == 0:
-            # apply Neumann condition at the origin
-            out_rr[i] = (arr_r[i + 1] - arr_r[i]) * scale_r
-            out_θθ[i] = arr_r[i] / rs[i]
-            out_φφ[i] = arr_r[i] / rs[i]
-        else:  # r_min > 0
-            out_rr[i] = (arr_r[i + 1] - value_r_lower_bc(arr_r, (i,))) * scale_r
-            out_θθ[i] = arr_r[i] / rs[i]
-            out_φφ[i] = arr_r[i] / rs[i]
-
-        for i in range(1, dim_r - 1):  # iterate radial points
+        for i in range(1, dim_r + 1):  # iterate radial points
             out_rr[i] = (arr_r[i + 1] - arr_r[i - 1]) * scale_r
-            out_θθ[i] = arr_r[i] / rs[i]
-            out_φφ[i] = arr_r[i] / rs[i]
-
-        # # outer radial boundary condition
-        i = dim_r - 1
-        out_rr[i] = (value_r_upper_bc(arr_r, (i,)) - arr_r[i - 1]) * scale_r
-        out_θθ[i] = arr_r[i] / rs[i]
-        out_φφ[i] = arr_r[i] / rs[i]
-
-        return out
+            out_θθ[i] = arr_r[i] / rs[i - 1]
+            out_φφ[i] = arr_r[i] / rs[i - 1]
 
     return vector_gradient  # type: ignore
 
 
 @SphericalSymGrid.register_operator("tensor_divergence", rank_in=2, rank_out=1)
 @fill_in_docstring
-def make_tensor_divergence(bcs: Boundaries, safe: bool = True) -> OperatorType:
+def make_tensor_divergence(grid: SphericalSymGrid, safe: bool = True) -> OperatorType:
     """make a discretized tensor divergence operator for a spherical grid
 
     {DESCR_SPHERICAL_GRID}
 
     Args:
-        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
-            {ARG_BOUNDARIES_INSTANCE}
+        grid (:class:`~pde.grids.spherical.SphericalSymGrid`):
+            The polar grid for which this operator will be defined
         safe (bool):
             Add extra checks for the validity of the input
 
     Returns:
         A function that can be applied to an array of values
     """
-    assert isinstance(bcs.grid, SphericalSymGrid)
-    bcs.check_value_rank(1)
+    assert isinstance(grid, SphericalSymGrid)
 
     # calculate preliminary quantities
-    dim_r = bcs.grid.shape[0]
-    r_min, _ = bcs.grid.axes_bounds[0]
-    rs = bcs.grid.axes_coords[0]
-    dr = bcs.grid.discretization[0]
+    dim_r = grid.shape[0]
+    rs = grid.axes_coords[0]
+    dr = grid.discretization[0]
     scale_r = 1 / (2 * dr)
 
-    # prepare boundary evaluators
-    bc_r = bcs.extract_component(0)[0]
-    value_r_lower_bc = bc_r.low.make_virtual_point_evaluator()
-    value_r_upper_bc = bc_r.high.make_virtual_point_evaluator()
-    bc_θ = bcs.extract_component(1)[0]
-    value_θ_lower_bc = bc_θ.low.make_virtual_point_evaluator()
-    value_θ_upper_bc = bc_θ.high.make_virtual_point_evaluator()
-    bc_φ = bcs.extract_component(2)[0]
-    value_φ_lower_bc = bc_φ.low.make_virtual_point_evaluator()
-    value_φ_upper_bc = bc_φ.high.make_virtual_point_evaluator()
-
-    @jit_allocate_out(out_shape=(3, dim_r))
-    def tensor_divergence(arr, out=None):
+    @jit
+    def tensor_divergence(arr: np.ndarray, out: np.ndarray) -> None:
         """apply tensor divergence operator to array `arr`"""
         # assign aliases
         arr_rr, arr_rθ, arr_rφ = arr[0, 0, :], arr[0, 1, :], arr[0, 2, :]
@@ -468,45 +297,22 @@ def make_tensor_divergence(bcs: Boundaries, safe: bool = True) -> OperatorType:
 
         # check inputs
         if safe:
-            assert np.all(arr_rθ == 0)
-            assert np.all(arr_θθ == 0)
-            assert np.all(arr_φφ == 0)
-            assert np.all(arr_φθ == 0)
-            assert np.all(arr_θφ == 0)
-
-        # evaluate innermost point
-        i = 0
-        if r_min == 0:
-            # apply Neumann condition at the origin
-            out_r[i] = (arr_rr[i + 1] - arr_rr[i]) * scale_r + 2 * arr_rr[i] / rs[i]
-            out_θ[i] = (arr_θr[i + 1] - arr_θr[i]) * scale_r + 2 * arr_θr[i] / rs[i]
-            term = (2 * arr_φr[i] + arr_rφ[i]) / rs[i]
-            out_φ[i] = (arr_φr[i + 1] - arr_φr[i]) * scale_r + term
-        else:  # r_min > 0
-            term = 2 * arr_rr[i] / rs[i]
-            out_r[i] = (arr_rr[i + 1] - value_r_lower_bc(arr_rr, (i,))) * scale_r + term
-            term = 2 * arr_θr[i] / rs[i]
-            out_θ[i] = (arr_θr[i + 1] - value_θ_lower_bc(arr_θr, (i,))) * scale_r + term
-            term = (2 * arr_φr[i] + arr_rφ[i]) / rs[i]
-            out_φ[i] = (arr_φr[i + 1] - value_φ_lower_bc(arr_φr, (i,))) * scale_r + term
+            assert np.all(arr_rθ[1:-1] == 0)
+            assert np.all(arr_θθ[1:-1] == 0)
+            assert np.all(arr_φφ[1:-1] == 0)
+            assert np.all(arr_φθ[1:-1] == 0)
+            assert np.all(arr_θφ[1:-1] == 0)
 
         # iterate over inner points
-        for i in range(1, dim_r - 1):
-            out_r[i] = (arr_rr[i + 1] - arr_rr[i - 1]) * scale_r + 2 * arr_rr[i] / rs[i]
-            out_θ[i] = (arr_θr[i + 1] - arr_θr[i - 1]) * scale_r + 2 * arr_θr[i] / rs[i]
-            term = (2 * arr_φr[i] + arr_rφ[i]) / rs[i]
-            out_φ[i] = (arr_φr[i + 1] - arr_φr[i - 1]) * scale_r + term
+        for i in range(1, dim_r + 1):
+            deriv_r = (arr_rr[i + 1] - arr_rr[i - 1]) * scale_r
+            out_r[i] = deriv_r + 2 * arr_rr[i] / rs[i - 1]
 
-        # evaluate outermost point
-        i = dim_r - 1
-        term = 2 * arr_rr[i] / rs[i]
-        out_r[i] = (value_r_upper_bc(arr_rr, (i,)) - arr_rr[i - 1]) * scale_r + term
-        term = 2 * arr_θr[i] / rs[i]
-        out_θ[i] = (value_θ_upper_bc(arr_θr, (i,)) - arr_θr[i - 1]) * scale_r + term
-        term = (2 * arr_φr[i] + arr_rφ[i]) / rs[i]
-        out_φ[i] = (value_φ_upper_bc(arr_φr, (i,)) - arr_φr[i - 1]) * scale_r + term
+            deriv_r = (arr_θr[i + 1] - arr_θr[i - 1]) * scale_r
+            out_θ[i] = deriv_r + 2 * arr_θr[i] / rs[i - 1]
 
-        return out
+            deriv_r = (arr_φr[i + 1] - arr_φr[i - 1]) * scale_r
+            out_φ[i] = deriv_r + (2 * arr_φr[i] + arr_rφ[i]) / rs[i - 1]
 
     return tensor_divergence  # type: ignore
 
