@@ -144,6 +144,7 @@ class ExpressionBase(metaclass=ABCMeta):
         self,
         expression: "basic.Basic",
         signature: Sequence[Union[str, List[str]]] = None,
+        *,
         user_funcs: Dict[str, Any] = None,
         consts: Dict[str, Any] = None,
     ):
@@ -381,6 +382,7 @@ class ExpressionBase(metaclass=ABCMeta):
         constants = tuple(self.consts)
 
         # turn the expression into a callable function
+        self._logger.info("Compile sympy expression `%s`", self._sympy_expr)
         func = sympy.lambdify(
             variables + constants,
             self._sympy_expr,
@@ -459,6 +461,7 @@ class ScalarExpression(ExpressionBase):
         self,
         expression: ExpressionType = 0,
         signature: Optional[Sequence[Union[str, List[str]]]] = None,
+        *,
         user_funcs: Optional[Dict[str, Any]] = None,
         consts: Optional[Dict[str, Any]] = None,
         allow_indexed: bool = False,
@@ -640,6 +643,7 @@ class TensorExpression(ExpressionBase):
         self,
         expression: ExpressionType,
         signature: Optional[Sequence[Union[str, List[str]]]] = None,
+        *,
         user_funcs: Optional[Dict[str, Any]] = None,
     ):
         """
